#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "asl.h"
#include "getstub.h"

/***
    
    Program purpose: to output analytical information about various
    problem characteristics.  Also, to optimize
    
    min || delf(x) - C(x)v || : 
    
    where x is the final solution of some optimization routine on
    some problem f(x) with constraints c(x) where C(x) is the 
    matrix of the gradients of constraints on f(x) in the original
    problem, including bound constraints.

***/

static double epstol = 1.0e-5;
static int    noopt  = 0;

static keyword keywds[] = {      /* must be in alphabetical order */
  KW("epstol", D_val, &epstol, "Tolerance for determining activity."),
  KW("noopt",  I_val, &noopt,  "Skip optimality calculation."),
};

static Option_Info Oinfo = {
  "analyze", "ANALYZE", "analyze_options", keywds, nkeywds
};

double delta(double x, double y)
{
  double abs_err = fabs(x - y);
  double rel_err = fabs(x) + fabs(y);

  if (rel_err > 0.0) {
    rel_err = abs_err / rel_err;
  }
  else {
    rel_err = abs_err;
  }

  if (abs_err < rel_err) {
    return abs_err;
  }
  return rel_err;
}

int main(int argc, char *argv[])
{
  char *home = getenv("ANALYZE_HOME");

  ASL *asl;
  FILE *nl, *fp, *fp2;
  char *stub;

  double *feaserr, *comperr;
  double *g, *c = NULL;
  int *v_active, *c_active = NULL;

  double err, lerr, uerr;
  double feas, comp;
  int i, j, active;

  int nvar_free = 0, nvar_fixed = 0, nvar_single = 0, nvar_double = 0;
  int ncon_unrestricted = 0, ncon_equality = 0;
  int ncon_inequality = 0, ncon_range = 0;

  char command[1024];

  asl = ASL_alloc(ASL_read_fg);		/* Function and gradients wanted     */
  asl->i.congrd_mode = 0; 		/* Dense Jacobian rows wanted        */
  want_xpi0 = 1;			/* Initial value for x wanted        */

  stub = getstub(&argv, &Oinfo); 	/* Obtain the name for the stub file */
  if (!stub) {
    usage_ASL(&Oinfo, 1);
  }

  nl = jac0dim(stub,(fint)strlen(stub));/* Read in first part of the stub    */
  if (getopts(argv, &Oinfo)) {		/* Set options provided by user      */
    exit(2);
  }
  fg_read(nl,0);			/* Read in the functions/Jacobian    */

  /**** Compute variable statistics ****/

  for (i = 0; i < n_var; ++i) {
    if ((LUv[2*i] == -Infinity) && (LUv[2*i+1] == Infinity)) {
      ++nvar_free;
    }
    else if (LUv[2*i] == LUv[2*i+1]) {
      ++nvar_fixed;
    }
    else if ((LUv[2*i] > -Infinity) && (LUv[2*i+1] < Infinity)) {
      ++nvar_double;
    }
    else {
      ++nvar_single;
    }
  }

  for (i = 0; i < n_con; ++i) {
    if ((LUrhs[2*i] == -Infinity) && (LUrhs[2*i+1] == Infinity)) {
      ++ncon_unrestricted;
    }
    else if (LUrhs[2*i] == LUrhs[2*i+1]) {
      ++ncon_equality;
    }
    else if ((LUrhs[2*i] > -Infinity) && (LUrhs[2*i+1] < Infinity)) {
      ++ncon_range;
    }
    else {
      ++ncon_inequality;
    }
  }

  /*** Print out model statistics ********************************************/
  printf("Number of variables: n = %d\n", n_var);
  printf("Number of binary variables: bn = %d\n", nbv);
  printf("Number of integer variables: in = %d\n", niv);
  printf("Number of free variables: %d\n", nvar_free);
  printf("Number of fixed variables: %d\n", nvar_fixed);
  printf("Number of single bound variables: %d\n", nvar_single);
  printf("Number of double bound variables: %d\n", nvar_double);
  printf("\n");

  printf("Number of constraints (not including bounds): c = %d\n", n_con);
  printf("Number of unrestricted constraints: %d\n", ncon_unrestricted);
  printf("Number of equality constraints: %d\n", ncon_equality);
  printf("Number of inequality constraints: %d\n", ncon_inequality);
  printf("Number of range constraints: %d\n", ncon_range);
  printf("\n");
 
  printf("Number of objectives: o = %d\n", n_obj);
  printf("Number of nonlinear objectives: nlo = %d\n", nlo);
  printf("Number of ranged constraints: rc = %d\n", nranges);
  printf("Number of nonlinear constraints: nlc = %d\n", nlc);
  printf("Number of nonlinear network constraints: nlnetc = %d\n", nlnc);
  printf("Number of nonlinear variables in both obj and "
         "constraints: nln = %d\n", nlvb);
  printf("Number of nonlinear integer variables in both: nlin = %d\n", nlvbi);
  printf("Number of nonlinear variables in constraints: nlnc = %d\n", nlvc);
  printf("Number of nonlinear integer variables in "
         "constraints: nlinc = %d\n", nlvci);
  printf("Number of nonlinear variables in objectives: nlno = %d\n", nlvo);
  printf("Number of nonlinear integer variables in "
         "objectives: nlino = %d\n", nlvoi);
  printf("Number of linear network constraints: lnetc = %d\n", lnc);
  printf("Number of nonzeros in Jacobian: nonzj = %d\n", nzc);
  printf("Number of nonzeros in objective gradients: nonzog = %d\n", nzo);
  fflush(stdout);

  /***
    Determine "activity" and violation of constraints.
    Equality constraints should lead to free multipliers,
    constraints close to their lower bounds should be
    bounded below, and constraints close to their upper
    bounds should be bounded above.
  ***/

  feaserr = (double *)malloc((n_var+n_con)*sizeof(double));
  comperr = (double *)malloc((n_var+n_con)*sizeof(double));

  g = (double *)malloc(n_var*sizeof(double));
  v_active = (int *)malloc(n_var*sizeof(int));

  if (n_con) {
    c = (double *)malloc(n_con*sizeof(double));
    c_active = (int *)malloc(n_con*sizeof(int));
    conval(X0, c, 0);
  }

  fp = fopen("bounds.dat", "w");
  fprintf(fp, "param : LambdaL LambdaU =\n");
  fprintf(fp, "0\t-Infinity\tInfinity\n");

  active = 0;
  for (i = 0; i < n_con; ++i) {
    /* Activity with respect to the lower bound on the constraint.           */
    if (LUrhs[2*i] == -Infinity) {
      lerr = Infinity;
    }
    else {
      lerr = delta(c[i], LUrhs[2*i]);
    }

    /* Activity with respect to the upper bound on the constraint.           */
    if (LUrhs[2*i+1] == Infinity) {
      uerr = Infinity;
    }
    else {
      uerr = delta(c[i], LUrhs[2*i+1]);
    }

    if (lerr < uerr) {
      err = lerr;
    }
    else {
      err = uerr;
    }

    if (c[i] >= LUrhs[2*i]) {
      if (c[i] <= LUrhs[2*i+1]) {
	feaserr[i] = 0.0;
      }
      else {
	feaserr[i] = uerr;
      }
    } 
    else {
      feaserr[i] = lerr;
    }
    
    if (err <= epstol) {
      c_active[i] = ++active;
      comperr[i] = err;      

      if ((lerr <= epstol) && (uerr <= epstol)) {
	fprintf(fp, "%d\t-Infinity\tInfinity\n", active);
      }
      else if (lerr <= epstol) {
	fprintf(fp, "%d\t0\tInfinity\n", active);
      }
      else if (uerr <= epstol) {
	fprintf(fp, "%d\t-Infinity\t0\n", active);
      }
    }
    else {
      c_active[i] = 0;
      comperr[i] = 0.0;
    }
  }

  for (i = 0; i < n_var; ++i) {
    /* Activity with respect to the lower bound on the variable.             */
    if (LUv[2*i] == -Infinity) {
      lerr = Infinity;
    }
    else {
      lerr = delta(X0[i], LUv[2*i]);
    }

    /* Activity with respect to the upper bound on the variable.             */
    if (LUv[2*i+1] == Infinity) {
      uerr = Infinity;
    }
    else {
      uerr = delta(X0[i], LUv[2*i+1]);
    }

    if (lerr < uerr) {
      err = lerr;
    }
    else {
      err = uerr;
    }

    if (X0[i] >= LUv[2*i]) {
      if (X0[i] <= LUv[2*i+1]) {
	feaserr[n_con+i] = 0.0;
      }
      else {
	feaserr[n_con+i] = uerr;
      }
    } 
    else {
      feaserr[n_con+i] = lerr;
    }

    if (err <= epstol) {
      v_active[i] = ++active;
      comperr[n_con+i] = err;      

      if ((lerr <= epstol) && (uerr <= epstol)) {
	fprintf(fp, "%d\t-Infinity\tInfinity\n", active);
      }
      else if (lerr <= epstol) {
	fprintf(fp, "%d\t0\tInfinity\n", active);
      }
      else if (uerr <= epstol) {
	fprintf(fp, "%d\t-Infinity\t0\n", active);
      }
    }
    else {
      v_active[i] = 0;
      comperr[n_con+i] = 0.0;
    }
  }
  fprintf(fp, ";\n");
  fclose(fp);

  feas = 0.0;
  comp = 0.0;
  for (i = 0; i < n_con + n_var; ++i) {
    if (feaserr[i] > feas) {
      feas = feaserr[i];
    }
    
    if (comperr[i] > comp) {
      comp = comperr[i];
    }
  }

  if (0 == noopt) {
    objgrd(0, X0, g, 0);

    fp = fopen("params.dat", "w");
    fprintf(fp, "param nvar := %d;\n", n_var);
    fprintf(fp, "param ncon := %d;\n", active);
    fprintf(fp, "param feas := %20.19e;\n", feas);
    fprintf(fp, "param comp := %20.19e;\n", comp);

    fprintf(fp, "\nparam gradf :=\n");

    if (!objtype[0]) {
      /* Minimize                                                            */
      for (i = 0; i < n_var; ++i) {
        if (LUv[2*i] != LUv[2*i+1]) {
          fprintf(fp, "%d %20.19e\n", i+1, g[i]);
        }
        else {
          fprintf(fp, "%d 0.0\n", i+1);
        }
      }
    }
    else {
      /* Maximize                                                            */
      for (i = 0; i < n_var; ++i) {
        if (LUv[2*i] != LUv[2*i+1]) {
          fprintf(fp, "%d %20.19e\n", i+1, -g[i]);
        }
        else {
          fprintf(fp, "%d 0.0\n", i+1);
        }
      }
    }

    fprintf(fp, ";\n");
    fclose(fp);

    fp = fopen("sparse.dat", "w");
    fprintf(fp, "set Nonzeros :=\n");

    fp2 = fopen("matrix.dat", "w");
    fprintf(fp2, "param J :=\n");

    for (i = 0; i < n_con; ++i) {
      if (c_active[i]) {
	congrd(i, X0, g,  0);      
	for (j = 0; j < n_var; j++) {
	  if ((0.0 != g[j]) && (LUv[2*j] != LUv[2*j+1])) {
	    fprintf(fp, "%d %d\n", j+1, c_active[i]);
	    fprintf(fp2, "%d %d %20.19e\n", j+1, c_active[i], g[j]);
	  }
	}
      }
    }

    for (i = 0; i < n_var; ++i) {
      if ((v_active[i]) && (LUv[2*i] != LUv[2*i+1])) {
        fprintf(fp, "%d %d\n", i+1, v_active[i]);
	fprintf(fp2, "%d %d 1.0\n", i+1, v_active[i]);
      }
    }

    fprintf(fp, ";\n");
    fprintf(fp2, ";\n");
    fclose(fp);
    fclose(fp2);

    strcpy(command, "ampl ");
    if (NULL != home) {
      strcat(command, getenv("ANALYZE_HOME"));
      strcat(command, "/check.com");
    } 
    else {
      strcat(command, "check.com");
    }

    system(command);

    unlink("params.dat");
    unlink("sparse.dat");
    unlink("matrix.dat");
  }
  else {
    printf("COPS feasibility error    : %e\n", feas);
    printf("COPS complementarity error: %e\n", comp);
    fflush(stdout);
  }

  unlink("bounds.dat");

  free(feaserr);
  free(comperr);

  free(g);
  free(v_active);

  if (n_con) {
    free(c);
    free(c_active);
  }

  solve_result_num = 0;
  write_sol("COPS COMPLETED", X0, pi0, 0);
  ASL_free(&asl);
  return 0;
}

