# Hang Glider Problem
# Trapezoidal formulation
# David Bortz - Summer 1998
# COPS 2.0 - September 2000
# COPS 3.0 - November 2002
# COPS 3.1 - March 2004

param x_0;   # Initial position and velocity of the glider.
param y_0;
param vx_0;
param vy_0;

param x_f;   # Final position and velocity of the glider.
param y_f;
param vx_f;
param vy_f;

param u_c;   # Parameters for the model.
param r_0;
param c0;
param c1;
param m;
param S;
param rho;
param g;

param nh;      # Time steps
var t_f;
var step = t_f/nh;

var x {0..nh};   # State variables
var y {0..nh};
var vx {0..nh};
var vy {0..nh};

var cL {0..nh};   # Control variable
param cL_min;           
param cL_max;

# Functions that define the glider.

var r {i in 0..nh} = (x[i]/r_0 - 2.5)^2;
var u {i in 0..nh} = u_c*(1-r[i])*exp(-r[i]);

var w {i in 0..nh} = vy[i] - u[i];
var v {i in 0..nh} = sqrt(vx[i]^2 + w[i]^2);

var D {i in 0..nh} = .5*(c0+c1*cL[i]^2)*rho*S*v[i]^2;
var L {i in 0..nh} = .5*cL[i]*rho*S*v[i]^2;       

# Functions in the equations of motion.

var vx_dot {i in 0..nh} = (-L[i]*(w[i]/v[i]) - D[i]*(vx[i]/v[i]))/m;
var vy_dot {i in 0..nh} = (L[i]*(vx[i]/v[i]) - D[i]*(w[i]/v[i]))/m - g;

maximize final_x: x[nh];

subject to step_eqn: step >= 0;

subject to cL_bounds {j in 0..nh}: cL_min <= cL[j] <= cL_max;
subject to x_bounds {j in 0..nh}: x[j] >= 0.0;
subject to vx_bounds {j in 0..nh}: 0.0 <= vx[j];

subject to x_eqn {j in 1..nh}:
   x[j] = x[j-1] + 0.5*step*(vx[j] + vx[j-1]);

subject to y_eqn {j in 1..nh}:
   y[j] = y[j-1] + 0.5*step*(vy[j] + vy[j-1]);

subject to vx_eqn {j in 1..nh}:
   vx[j] = vx[j-1] + 0.5*step*(vx_dot[j] + vx_dot[j-1]);

subject to vy_eqn {j in 1..nh}:
   vy[j] = vy[j-1] + 0.5*step*(vy_dot[j] + vy_dot[j-1]);

# Boundary Conditions

subject to x_ic : x[0] = x_0;
subject to y_ic : y[0] = y_0;
subject to y_fc : y[nh] = y_f;
subject to vx_ic : vx[0] = vx_0;
subject to vx_fc : vx[nh] = vx_f;
subject to vy_ic : vy[0] = vy_0;
subject to vy_fc : vy[nh] = vy_f;

