# Todd Munson
# COPS 3.1 - March 2004

###############################################################################
# Generalized Lane-Emden-Fowler-Henon equation with homogeneous Dirichlet 
# boundary conditions:
#
#   a*Laplacian - b(x)*u + c(x)*u - d(x) = 0	for x in  int Omega
#   u(x) = 0					for x in bdry Omega
#
# Equation classes considered:
#   Lane-Emden-Fowler: c(x) = 1
#   Henon            : c(x) = |x|^(2*l) 
###############################################################################

###############################################################################
# Description of the discretization
###############################################################################

param NODES;				# Number of nodes
param ELEMS;				# Number of elements
param DIMEN := 2;			# Number of spatial dimensions

set D := {1..DIMEN};			# Set for the dimensions
set N := {1..NODES};			# Set for the nodes
set E := {1..ELEMS};			# Set for the elements

param TRIANG{E, 1..DIMEN+1} in N;	# Description of triangular elements
param COORDS{N, D};			# Coordinates for the nodes
param BNDRY{N} binary;			# Indicator for the boundary nodes

param EDGE{e in E, d1 in 1..DIMEN+1, d2 in D} := 	# Edge lengths
  COORDS[TRIANG[e,(d1 mod (DIMEN+1)) + 1],d2] - COORDS[TRIANG[e,d1],d2];

param AREA{e in E} := 					# Area of element
  (EDGE[e,1,1]*EDGE[e,2,2] - EDGE[e,1,2]*EDGE[e,2,1]) / 2;

check {e in E}: AREA[e] > 0;

###############################################################################
# Description of the partial differential equation
###############################################################################

param a    >  0;			# Coefficient for laplacian
param b{N} >= 0;			# Coefficient for u
param c{N} >= 0;			# Coefficient for u^p
param p{N} >  0;			# Power
param d{N};				# Right-hand side

check {n in N: BNDRY[n]}: d[n] = 0;

###############################################################################
# Mountain pass algorithm information
#   BREAK -- number of breakpoints for path generated
#   US    -- starting point
#   UE    -- ending point
#   ALPHA -- constant for a bound on the length of the path generated in
#            relation to the distance between the starting and ending point
###############################################################################

param BREAK;
param US{N};
param UE{N};
param ALPHA >= 1;

# Check to make sure the boundary conditions are correctly set
check {n in N: BNDRY[n]}: ((US[n] = 0) and (UE[n] = 0));

# Derived quantities

param H := ALPHA/(BREAK+1)*sqrt(sum {n in N} (US[n]-UE[n])^2);
check: H > 0;

###############################################################################
# Optimization problem specification
###############################################################################

var u{0..BREAK+1, N};
var z;

###############################################################################
# Assume right triangles with right angle at node 1 of the triangle.
# Split into cases depending upon whether the x coordinate of edge 1 is
# positive or the y coordinate of edge 1 is positive.
###############################################################################

var integral{b1 in 0..BREAK+1, e1 in E} =
AREA[e1]*(
  1 / (DIMEN+1) *
    (sum {c1 in 1..DIMEN+1} (b[TRIANG[e1,c1]]*u[b1,TRIANG[e1,c1]]^2/2-
			     c[TRIANG[e1,c1]]*u[b1,TRIANG[e1,c1]]^(p[TRIANG[e1,c1]]+1)/(p[TRIANG[e1,c1]]+1)+
 			     d[TRIANG[e1,c1]]*u[b1,TRIANG[e1,c1]])) +
  a / (8*AREA[e1]^2)*(
      u[b1,TRIANG[e1,1]]^2*(EDGE[e1,2,1]^2 + EDGE[e1,2,2]^2) +
      u[b1,TRIANG[e1,2]]^2*(EDGE[e1,3,1]^2 + EDGE[e1,3,2]^2) +
      u[b1,TRIANG[e1,3]]^2*(EDGE[e1,1,1]^2 + EDGE[e1,1,2]^2) +
    2*u[b1,TRIANG[e1,1]]*u[b1,TRIANG[e1,2]]*(EDGE[e1,2,1]*EDGE[e1,3,1] + EDGE[e1,2,2]*EDGE[e1,3,2]) +
    2*u[b1,TRIANG[e1,1]]*u[b1,TRIANG[e1,3]]*(EDGE[e1,2,1]*EDGE[e1,1,1] + EDGE[e1,2,2]*EDGE[e1,1,2]) +
    2*u[b1,TRIANG[e1,2]]*u[b1,TRIANG[e1,3]]*(EDGE[e1,1,1]*EDGE[e1,3,1] + EDGE[e1,1,2]*EDGE[e1,3,2])
  ));

var energy{b1 in 0..BREAK+1} = sum {e1 in E} integral[b1,e1];

minimize cost: z;

subject to max_energy {b1 in 1..BREAK}: 
  z >= energy[b1];

subject to distance {b1 in 0..BREAK}:
  sum {n in N} (u[b1+1,n]-u[b1,n])^2 <= H^2;

subject to boundary {b1 in 1..BREAK, n in N: BNDRY[n]}:
  u[b1,n] = 0;

subject to start {n in N}:
  u[0,n] = US[n];

subject to end {n in N}:
  u[BREAK+1,n] = UE[n];

