# Robot Arm Problem
# Trapezoidal formulation
# David Bortz - Summer 1998
# COPS 2.0 - September 2000
# COPS 3.0 - November 2002
# COPS 3.1 - March 2004

param nh;               # number of intervals.
param L;		# total length of arm
param pi := 4*atan(1);

# Upper bounds on the controls

param max_u_rho;
param max_u_the;
param max_u_phi;

# Initial positions of the length and the angles for the robot arm

param rho_0;
param the_0;
param phi_0;

# Final positions of the length and the angles for the robot arm

param rho_n;
param the_n;
param phi_n;

# The length and the angles theta and phi for the robot arm.

var rho {0..nh} >=0, <= L; 
var the {0..nh} >= -pi, <= pi; 
var phi {0..nh} >=0, <= pi;

# The derivatives of the length and the angles.
    
var rho_dot {0..nh};
var the_dot {0..nh};
var phi_dot {0..nh};

# The controls.

var u_rho {0..nh} >= -max_u_rho, <= max_u_rho;
var u_the {0..nh} >= -max_u_the, <= max_u_the;
var u_phi {0..nh} >= -max_u_phi, <= max_u_phi;

# The step and the final time.

var tf;
var step = tf/nh;

# The moments of inertia.

var I_the {i in 0..nh} = ((L-rho[i])^3+rho[i]^3)*(sin(phi[i]))^2/3.0;
var I_phi {i in 0..nh} = ((L-rho[i])^3+rho[i]^3)/3.0;

# The robot arm problem.

minimize time: tf;

subject to rho_eqn {j in 1..nh}:
        rho[j] = rho[j-1] + 0.5*step*(rho_dot[j] + rho_dot[j-1]);

subject to the_eqn {j in 1..nh}:
        the[j] = the[j-1] + 0.5*step*(the_dot[j] + the_dot[j-1]);

subject to phi_eqn {j in 1..nh}:
        phi[j] = phi[j-1] + 0.5*step*(phi_dot[j] + phi_dot[j-1]);

subject to u_rho_eqn {j in 1..nh}:
        rho_dot[j] = rho_dot[j-1] + 0.5*step*(u_rho[j] + u_rho[j-1])/L;

subject to u_the_eqn {j in 1..nh}:
        the_dot[j] = the_dot[j-1] + 
                   0.5*step*(u_the[j]/I_the[j] + u_the[j-1]/I_the[j-1]);

subject to u_phi_eqn {j in 1..nh}:
        phi_dot[j] = phi_dot[j-1] + 
                   0.5*step*(u_phi[j]/I_phi[j] + u_phi[j-1]/I_phi[j-1]);

# Boundary Conditions

subject to rho_0_eqn: rho[0] = 4.5;
subject to the_0_eqn: the[0] = 0;
subject to phi_0_eqn: phi[0] = pi/4;

subject to rho_f_eqn: rho[nh] = 4.5;
subject to the_f_eqn: the[nh] = 2*pi/3;
subject to phi_f_eqn: phi[nh] = pi/4;

subject to rho_dot_0_eqn: rho_dot[0] = 0;
subject to the_dot_0_eqn: the_dot[0] = 0;
subject to phi_dot_0_eqn: phi_dot[0] = 0;

subject to rho_dot_f_eqn: rho_dot[nh] = 0;
subject to the_dot_f_eqn: the_dot[nh] = 0;
subject to phi_dot_f_eqn: phi_dot[nh] = 0;

